from dataclasses import dataclass
from typing import Union, Callable, Generic, TypeVar


P = TypeVar("P")
@dataclass
class Success(Generic[P]):
    value: P
    next_index: int


@dataclass
class Failure:
    at_index: int
    expected: set[str]
    depth: int | None


P = TypeVar("P")
Result = Union["Failure", "Success[P]"]


class ParseError(BaseException):
    failure: Failure
    stream: str

    def __init__(self, failure: Failure, stream: str) -> None:
        self.failure = failure
        self.stream = stream

        fail_section = stream[failure.at_index:80]
        message = f"""
Parsing failed at position {self.failure.at_index} of stream :
${fail_section}

Expected one of:
{failure.expected}
"""
        super().__init__(message)
